#include <string.h>

#include <qfile.h>
#include <qfileinfo.h>
#include <qmap.h>
#include <qstringlist.h>
#include <qtextstream.h>

#include "configfile.h"

ConfigFile::ConfigFile(const QString &filename)
           :filename(filename)
{
  if (filename.length() > 0)
  {
    git = groups.end();
    read();
  }
}

ConfigFile::~ConfigFile()
{
  if (filename.length() > 0)
    write();
}

void ConfigFile::setGroup(const QString &gname)
{
  QMap< QString, ConfigGroup>::Iterator it = groups.find(gname);
  if (it == groups.end())
  {
    ConfigGroup *grp = new ConfigGroup;
    git = groups.insert(gname, *grp);
    return;
  }
  git = it;
}

void ConfigFile::writeEntry(const QString &key, const QString &value)
{
  if (git == groups.end()) {
    qWarning("no group set");
    return;
  }
  (*git ).insert(key, value);
}

void ConfigFile::writeEncryptedEntry(const QString &key, const QString &value)
{
  if (git == groups.end()) {
    qWarning("no group set");
    return;
  }

  char encrypted[1024];
  strcpy(encrypted, (value + value).latin1());
  for (unsigned i=0; i<(value.length() * 2); i++)
  {
    int chr = encrypted[i];
    chr += 6 + 4 * i;
    while (chr > 126) chr -= 94;
    encrypted[i] = chr;
  }

  (*git ).insert(key, QString(encrypted));
}

void ConfigFile::writeEntry(const QString &key, int num)
{
  QString s;
  s.setNum(num);
  writeEntry(key, s);
}

void ConfigFile::writeEntry(const QString &key, bool b)
{
  QString s;
  s.setNum((int) b);
  writeEntry(key, s);
}

void ConfigFile::writeEntry(const QString &key, const QStringList &lst, const QChar &sep)
{
  QString s;
  QStringList::ConstIterator it = lst.begin();

  for (; it != lst.end(); ++it)
    s += *it + sep;
  writeEntry(key, s);
}

QString ConfigFile::readEntry(const QString &key, const QString &deflt)
{
  if (git == groups.end()) {
    qWarning("no group set");
    return deflt;
  }
  ConfigGroup::Iterator it = (*git).find(key);
  if (it != (*git).end())
    return *it;
  else
    return deflt;
}

QString ConfigFile::readEncryptedEntry(const QString &key, const QString &deflt)
{
  if (git == groups.end()) {
    qWarning("no group set");
    return deflt;
  }
  ConfigGroup::Iterator it = (*git).find(key);
  if (it != (*git).end())
  {
    char decrypted[1024];
    strcpy(decrypted, (*it).latin1());
    decrypted[(*it).length() / 2] = 0;
    for (unsigned i=0; i<((*it).length() / 2); i++)
    {
      int chr = decrypted[i];
      chr -= 6 + 4 * i;
      while (chr < 33) chr += 94;
      decrypted[i] = chr;
    }

    return QString(decrypted);
  }
  else
    return deflt;
}

int ConfigFile::readNumEntry(const QString &key, int deflt)
{
  QString s = readEntry(key);
  if (s.isEmpty())
    return deflt;
  else
    return s.toInt();
}

bool ConfigFile::readBoolEntry(const QString &key, bool deflt)
{
  QString s = readEntry(key);
  if (s.isEmpty())
    return deflt;
  else
    return (bool)s.toInt();
}

QStringList ConfigFile::readListEntry(const QString &key, const QChar &sep)
{
  QString s = readEntry(key);
  if (s.isEmpty())
    return QStringList();
  else
    return QStringList::split(sep, s);
}

void ConfigFile::clearGroup()
{
  if (git == groups.end()) {
    qWarning("no group set");
    return;
  }
  (*git).clear();
}

void ConfigFile::deleteGroup()
{
  if (git == groups.end()) {
    qWarning("no group set");
    return;
  }
  #if QT_VERSION >= 300
    groups.erase(git);
  #endif
  git = groups.end();
}

status_t ConfigFile::write(const QString &fn)
{
  if (!fn.isEmpty())
  filename = fn;

  QFile f(filename);
  if (!f.open(IO_WriteOnly)) {
    qWarning("could not open for writing `%s'", filename.latin1());
    git = groups.end();
    return T_ERROR;
  }

  QTextStream s(&f);
  QMap<QString, ConfigGroup>::Iterator g_it = groups.begin();
  for (; g_it != groups.end(); ++g_it)
  {
    s << "[" << g_it.key() << "]" << "\n";
    ConfigGroup::Iterator e_it = (*g_it).begin();

    for ( ; e_it != (*g_it).end(); ++e_it)
      s << e_it.key() << " = " << *e_it << "\n";
  }

  f.close();

  return T_OK;
}

void ConfigFile::read()
{
  if (!QFileInfo(filename).exists()) {
    qWarning("`%s' doesn't exist", filename.latin1());
    git = groups.end();
    return;
  }

  QFile f(filename);
  if (!f.open(IO_ReadOnly)) {
    qWarning("could not open for reading `%s'", filename.latin1());
    git = groups.end();
    return;
  }

  QTextStream s(&f);
  QString line;
  while (!s.atEnd()) {
    line = s.readLine();
    parse(line);
  }
  f.close();
}

void ConfigFile::parse(const QString &l)
{
  QString line = l.stripWhiteSpace();

  if (line[ 0 ] == QChar('[')) {
    QString gname = line;
    gname = gname.remove(0, 1);

    if (gname[ (int)gname.length() - 1 ] == QChar(']'))
      gname = gname.remove(gname.length() - 1, 1);
    ConfigGroup *grp = new ConfigGroup;
    git = groups.insert(gname, *grp);
  } else {

    if (git == groups.end()) {
      qWarning("line `%s' out of group", line.latin1());
      return;
    }

    QStringList lst = QStringList::split('=', line);
    if (lst.count() != 2 && line.find('=') == -1) {
      qWarning("corrupted line `%s' in group `%s'", line.latin1(), git.key().latin1());
      return;
    }

    (*git).insert(lst[ 0 ].stripWhiteSpace(), lst[1].stripWhiteSpace());
  }
}
