#include <iostream>

#include "mpdcom.h"

#define POLL_TIME 1900

/*! \param hostname             the host to connect to
    \param hostport             the port on the host to connect to
    \param needsPlayList        whether an automatic update of the playlist is needed
    \param parent               a parent, defaults to NULL
*/
MpdCom::MpdCom( const char * hostname, int hostport, bool needsPlayList, QObject * parent = 0 )
       :QObject(parent),
        commandQueue(),
        hostname(hostname),
        hostport(hostport),
        needsPlayList(needsPlayList),
        socket(this),
        statusTimer(this),
        reconnectTimer(this),
        oldvol(-1),
        buildingCommandList(false)
{
  dbgConnect( &socket, SIGNAL(readyRead()), this, SLOT(slotReadyRead()) );
  dbgConnect( &socket, SIGNAL(connected()), this, SLOT(slotConnected()) );
  dbgConnect( &socket, SIGNAL(error(int)), this, SLOT(slotError(int)) );
  dbgConnect( &socket, SIGNAL(connectionClosed()), this, SLOT(slotConnectionClosed()) );
  dbgConnect( &statusTimer, SIGNAL(timeout()), this, SLOT(requestStatusUpdate()) );
  dbgConnect( &reconnectTimer, SIGNAL(timeout()), this, SLOT(slotReconnect()) );
}

/*! \param hostname             the new hostname
    \param hostport             the new port on the host
*/
void MpdCom::reconnectCom(const char *hostname, int hostport)
{
  this->hostname = hostname;
  this->hostport = hostport;
  bool wasConnected = socket.state() != QSocket::Idle;

  disconnectCom();
  if (wasConnected)
    connectCom();
}

void MpdCom::disconnectCom(void)
{
  debug( "Disconnecting ... " );
  statusTimer.stop();
  socket.close();
  while( !commandQueue.empty() ) commandQueue.pop();
}

void MpdCom::connectCom(void)
{
  debug( "Connecting ... " );
  commandQueue.push( new CommandItem( "", CONNECTING, true ) );
  socket.connectToHost(hostname, hostport);
  statusTimer.start(POLL_TIME);
  requestStatusUpdate();
}

void MpdCom::previous(void)
{
  postCommand("previous");
  requestStatusUpdate();
}

void MpdCom::next(void)
{
  postCommand("next");
  requestStatusUpdate();
}

void MpdCom::stop(void)
{
  postCommand("stop");
  requestStatusUpdate();
}

void MpdCom::play(void)
{
  postCommand("play");
  requestStatusUpdate();
}

/*! \param songnr               the song to play
*/
void MpdCom::play(int songnr)
{
  postCommand("play " + QString::number(songnr));
  requestStatusUpdate();
}

void MpdCom::pause(void)
{
  postCommand("pause");
  requestStatusUpdate();
}

/*! \param songnr               the song to play
    \param pos                  the position in the song
*/
void MpdCom::seek(int songnr, int pos)
{
  postCommand("seek " + QString::number(songnr) + " " + QString::number(pos));
  requestStatusUpdate();
}

/*! This function is version independant, MPD versions below 0.10 work with
    relative values. The absolute values of this function are translated for older versions.
    \param volume               the absolute value for the volume ranging between 0 and 100
*/
void MpdCom::volume(int perc)
{
  if(minVersion("0.10")) postCommand("setvol " + QString::number(perc));
  else
  {
    postCommand("volume " + QString::number(perc-oldvol));
  }
}

void MpdCom::repeat(bool on)
{
  postCommand("repeat " + QString::number(on ? 1 : 0));
}

void MpdCom::random(bool on)
{
  postCommand("random " + QString::number(on ? 1 : 0));
}

void MpdCom::update(void)
{
  postCommand("update");
}

/*! Only works on MPD 0.10+
    \param sec                  the amount of seconds to crossfade
*/
void MpdCom::crossFade(int sec)
{
  if(minVersion("0.10"))
    postCommand("crossfade " + QString::number(sec));
  else
    debug("Crossfade not supported in this version, please upgrade to MPD 0.10+");
}

void MpdCom::requestPlaylist(void)
{
  for(Playlist::iterator i=playlist.begin(); i!=playlist.end(); i++)
    if(*i) delete *i;
  playlist.clear();
  songIndex = 0;
  si = new Playlist::SongInfo();
  postCommand("playlistinfo", PLAYLIST);
}

/*! \param playlistname          filename of the playlist without extension
    \param enqueue               if true the current playlist is not cleared
   */
void MpdCom::loadPlaylist(QString playlistname, bool enqueue)
{
  if( !enqueue ) postCommand("clear");
  postCommand("load " + playlistname);
  if(!enqueue) play();
}

void MpdCom::clearPlaylist(void)
{
  postCommand("clear");
}

/*! \param playlist             the playlist to remove
*/
void MpdCom::deletePlaylist(QString playlist)
{
  postCommand("rm " + playlist);
}

/*! \param index                the index of the song
*/
void MpdCom::requestSongInfo(unsigned index)
{
  songInfo.clear(index); postCommand("playlistinfo " + QString::number(index), PLAYLIST_INFO);
}

void MpdCom::requestStatusUpdate(void)
{
  postCommand("status", STATUS);
}

/*! \param dir                  the target directory
*/
void MpdCom::requestDirectoryContent(QString dir)
{
  debug( "MpdCom::requestDirectoryContent: " << dir.latin1() );
  filelist.clear();
  if(dir.length() != 0) dir = "\"" + dir + "\"";
  postCommand("lsinfo " + dir, FILELIST);
}

/*! Directories are only supported from MPD 0.10+.
    \param name                 the file/dir to add
*/
void MpdCom::addFile( QString name )
{
  postCommand("add \"" + name + "\"");
}

/*! \param index                the song to remove
*/
void MpdCom::removeFile( QString index )
{
  postCommand("delete " + index);
}

/*! \param index1               the index of the first song
    \param index2               the index of the second song
*/
void MpdCom::swap(unsigned index1, unsigned index2)
{
  postCommand("swap " + QString::number(index1) + " " + QString::number(index2));
}

/*! \param ver                  the version
*/
void MpdCom::setVersion(QString ver)
{
  majorVersion = minorVersion = buildVersion = 0;
  int i = ver.find('.');
  if( i > 0 )
  {
    majorVersion = ver.left(i).toInt();
    ver = ver.mid(i+1);
    i = ver.find('.');
    if( i > 0 )
    {
      minorVersion = ver.left(i).toInt();
      ver = ver.mid(i+1);
      buildVersion = ver.toInt();
    }
    else minorVersion = ver.toInt();
  }
  else majorVersion = ver.toInt();
}

/*! \param ver                  the version to compare with
*/
bool MpdCom::minVersion(QString ver)
{
  unsigned major = 0, minor = 0, build = 0;
  int i = ver.find('.');
  if( i > 0 )
  {
    major = ver.left(i).toInt();
    ver = ver.mid(i+1);
    i = ver.find('.');
    if( i > 0 )
    {
      minor = ver.left(i).toInt();
      ver = ver.mid(i+1);
      build = ver.toInt();
    }
    else minor = ver.toInt();
  }
  else major = ver.toInt();

  if( majorVersion >= major && minorVersion >= minor && buildVersion >= build ) return true;
  return false;
}

/*! \param data                 the command to send
    \param responce             the expected responce
*/
void MpdCom::postCommand( QString data, responcetype responce )
{
  if(buildingCommandList && responce == SIMPLE)
    commandList += data + "\n";
  else
  {
    commandQueue.push( new CommandItem( data+"\n", responce ) );
    if(commandQueue.size() == 1) sendData();
  }
}

void MpdCom::sendData(void)
{
  CommandItem * item = commandQueue.front();
  test( !item, "Invalid commandstructure" );
  if( !item->send )
  {
    if( socket.writeBlock(item->command.latin1(), item->command.length()) == -1 )
      test(true, "transmit failed");
    item->send = true;
  }
}

void MpdCom::slotReadyRead(void)
{
  QString str = "";
  if( commandQueue.size() == 0 )
  {
    //Flush data
    while( socket.canReadLine() ) str = socket.readLine();
    //Assert
    test(true, "Unxpected data" );
  }
  CommandItem * item = commandQueue.front();
  test( !item, "Invalid commandstructure" );
  test( item->send==false, "Unexpected data" );
  switch( item->responce )
  {
    case CONNECTING :
      str = socket.readLine();
      if( str.left(2) == "OK" ){
        setVersion( str.mid(7) );
        debug( str.mid(3).latin1() );
      }
      else test( true, (QString("Connection failure: ") + str).latin1() );
      commandQueue.pop();
      break;
    case SIMPLE :
      str = socket.readLine();
      if( str.left(2) != "OK" ) test( true, "Command error: " + str );
      commandQueue.pop();
      break;
    case PLAYLIST:
    {
      while( socket.canReadLine() )
      {
        str = socket.readLine();
        str.truncate(str.length()-1);
        if( str.startsWith("file:") || str.startsWith("OK")  )
        {
          if( songIndex > 0 )
          {
            if(si->title.length()==0)
              if( !getSongArtist(si->filename, si->artist, si->title) ) si->title = si->filename;

            si->songIndex = songIndex-1;
            playlist.addPlaylistItem( si );
          }
          if(str.startsWith("file:"))
          {
            si = new Playlist::SongInfo();
            songIndex++;
          }
          else
          {
            debug("playlist length: " << playlist.size());
            playlistSize = playlist.size();
            emit playlistUpdate( playlist );
            commandQueue.pop();
          }
          si->filename   = str.mid(6);
        }
        else if( str.startsWith("Artist:") )  si->artist     = str.mid(8);
        else if( str.startsWith("Album:")  )  si->album      = str.mid(7);
        else if( str.startsWith("Title:")  )  si->title      = str.mid(7);
        else if( str.startsWith("Time:")   )  si->songLength = str.right(str.length()-6).toInt();
      }
      break;
    }
    case PLAYLIST_INFO:
    {
      while( socket.canReadLine() )
      {
        str = socket.readLine();
        str.truncate(str.length()-1);
        if     ( str.startsWith("file:")   )  songInfo.filename   = str.mid(6);
        else if( str.startsWith("Artist:") )  songInfo.artist     = str.mid(8);
        else if( str.startsWith("Album:")  )  songInfo.album      = str.mid(7);
        else if( str.startsWith("Title:")  )  songInfo.title      = str.mid(7);
        else if( str.startsWith("Time:")   )  songInfo.songLength = str.mid(6).toInt();
        else if( str.startsWith("OK") )
        {
          commandQueue.pop();
//          requestStatusUpdate();
        }
      }
      break;
    }
    case STATUS:
    {
      Status status;
      if(!minVersion("0.10")) status.xfade = 0;
      while( socket.canReadLine() )
      {
        str = socket.readLine();
        str.truncate(str.length()-1);
        if     ( str.startsWith("volume:" ) ) status.volume = str.mid(8).toUInt();
        else if( str.startsWith("repeat:" ) ) status.repeat = str.right(1) == "0" ? false : true;
        else if( str.startsWith("random:" ) ) status.random = str.right(1) == "0" ? false : true;
        else if( str.startsWith("playlist:" ) ) status.playlist = str.mid(10).toUInt();
        else if( str.startsWith("playlistlength:" ) ) status.playlistLength = str.mid(16).toInt() * 10;
        else if( str.startsWith("state:" ) ) status.state = str.right(4) == "play" ? PLAY : ( str.right(4) == "stop" ? STOP : PAUSE );
        else if( str.startsWith("song:" ) ) status.songIndex = status.songIndex = str.mid(6).toUInt();
        else if( str.startsWith("time:" ) )
        {
          str = str.mid(6);
          int i = str.find(':');
          status.elapsed = str.left(i).toInt() * 10;
          status.songLength = str.right(str.length()-i-1).toInt() * 10;
        }
        else if( str.startsWith("bitrate:") ) status.bitrate = str.mid(9).toInt();
        else if( str.startsWith("xfade:")   ) status.xfade = str.mid(7).toInt();
        else if( str.startsWith("audio:") )
        {
          str = str.mid(7);
          int i = str.find(':');
          int j = str.findRev(':');

          status.samplerate = str.left(i).toInt();
          status.bits = str.mid(i+1,j-i-1).toInt();
          status.channels = str.mid(j+1).toInt();
        }
        else if( str.left(2) == "OK" )
        {
          oldvol = status.volume;
          commandQueue.pop();
          if( status.songIndex != songInfo.songIndex )
          {
            status.songInfo = NULL;
            requestSongInfo(status.songIndex);
          }
          else status.songInfo = &songInfo;
          if( currentPlaylist != status.playlist )
          {
            currentPlaylist = status.playlist;
            if (needsPlayList)
              requestPlaylist();
          }
          emit statusUpdate(status);
        }
        else debug(str.latin1());
      }
      break;
    }
    case FILELIST:
      debug( "receiving filelist" );
      while( socket.canReadLine() )
      {
        str = socket.readLine();
        str.truncate(str.length()-1);

        if     ( str.startsWith("directory:") ) filelist.add(str.mid(11), Filelist::DIR);
        else if( str.startsWith("file:" ) ) filelist.add(str.mid(6), Filelist::FILE);
        else if( str.startsWith("playlist:" ) ) filelist.add(str.mid(10), Filelist::PLAYLIST);
        else if( str.startsWith("OK") )
        {
          commandQueue.pop();
          emit directoryContent(filelist);
        }
      }
      break;
    default:
      emit error("Unknown responce type");
      break;
  }
  if( str.startsWith("ACK") )
  {
    emit error("Server returned an error: " + str.mid(5));
    commandQueue.pop();
  }

  if( commandQueue.size() > 0 ) sendData();
}

void MpdCom::slotConnected(void)
{
  debug( "Connection established" );
}

/*! \param e                    the error number
*/
void MpdCom::slotError(int e)
{
  switch(e)
  {
    case QSocket::ErrConnectionRefused : emit error("Connection refused"); break;
    case QSocket::ErrHostNotFound      : emit error("Host not found"); break;
    case QSocket::ErrSocketRead        : emit error("Socket read error"); break;
  }
  reconnectTimer.start(5000, true);
}

void MpdCom::flushCommandQueue(void)
{
  while(commandQueue.size() > 0) commandQueue.pop();
}

void MpdCom::slotReconnect(void)
{
  if( socket.state() == QSocket::Idle )
  {
    while(commandQueue.size() > 0) commandQueue.pop();
    commandQueue.push( new CommandItem( "", CONNECTING, true ) );
    socket.connectToHost(hostname, hostport);
  }
}

void MpdCom::slotConnectionClosed(void)
{
  debug( "Connection closed reconnecting..." );
  slotReconnect();
}
