#include "playlistmanager.h"
#include "configuration.h"

//Static init
PlaylistManager * PlaylistManager::me = NULL;

/*! \param mpdCom               MPD connection
*/
PlaylistManager::PlaylistManager(MpdCom * mpdCom)
                :QDialog(0),
                 currDir(""),
                 mpdCom(mpdCom),
                 mainLayout(this),
                 buttonLayout(this),
                 leftLayout(this),
                 rightLayout(this),
                 lvPlaylists(this),
                 lvDirectories(this),
                 lvFiles(this),
                 btAdd(LANG_ADD_SELECTION, this),
                 btUpdate(LANG_UPDATE, this)
{
  setGeometry(Configuration::getGeometry("playlistmanager"));

  lvPlaylists.addColumn(LANG_PLAYLISTS);
  lvDirectories.addColumn(LANG_DIRECTORIES);
  lvFiles.addColumn(LANG_FILES);
  lvFiles.setSelectionMode(QListView::Extended);

  lvPlaylists.setColumnWidthMode (0, QListView::Manual);
  lvDirectories.setColumnWidthMode (0, QListView::Manual);
  lvFiles.setColumnWidthMode (0, QListView::Manual);

  leftLayout.addWidget(&lvPlaylists, 1);
  leftLayout.addWidget(&lvDirectories, 2);

  buttonLayout.addWidget(&btAdd);
  buttonLayout.addStretch();
  buttonLayout.addWidget(&btUpdate);
  rightLayout.addWidget(&lvFiles, 2);
  rightLayout.addLayout(&buttonLayout, 0);

  mainLayout.addLayout(&leftLayout, 1);
  mainLayout.addLayout(&rightLayout, 2);

  dbgConnect(mpdCom, SIGNAL(directoryContent(const MpdCom::Filelist &)), this, SLOT(slotReceiveDirectoryContent(const MpdCom::Filelist &)));
  dbgConnect(&lvDirectories, SIGNAL(doubleClicked(QListViewItem *)), this, SLOT(slotDirectoryDoubleClicked(QListViewItem *)));
  dbgConnect(&lvDirectories, SIGNAL(rightButtonPressed(QListViewItem *, const QPoint &, int)), this, SLOT(slotDirectoryRightClick(QListViewItem *, const QPoint &, int)));
  dbgConnect(&lvFiles, SIGNAL(doubleClicked(QListViewItem *)), this, SLOT(slotFileDoubleClicked(QListViewItem *)));
  dbgConnect(&lvPlaylists, SIGNAL(doubleClicked(QListViewItem *)), this, SLOT(slotPlaylistDoubleClicked(QListViewItem *)));
  dbgConnect(&lvPlaylists, SIGNAL(rightButtonPressed(QListViewItem *, const QPoint &, int)), this, SLOT(slotPlaylistRightClick(QListViewItem *, const QPoint &, int)));
  dbgConnect(&btUpdate, SIGNAL(clicked()), mpdCom, SLOT(update()));
  dbgConnect(&btAdd, SIGNAL(clicked()), this, SLOT(slotAddSelection()));

  mpdCom->requestDirectoryContent();
}

/*! Use this function to show the PlaylistManager. It constructs a window when used the first time.
    After the first time the window is reused for better performance.
    \param mpdCom               the MpdCom class to use for communication
    \param directPlay           if true the playlist is cleared, the selection is added and the window is closed.
*/
void PlaylistManager::showme(MpdCom * mpdCom, bool directPlay)
{
  if(!me) me = new PlaylistManager(mpdCom);
  if(me->mpdCom != mpdCom) me->mpdCom = mpdCom;
  me->directPlay = directPlay;
  me->exec();
}

/*! \param filelist             the filelist received from MPD.
*/
void PlaylistManager::slotReceiveDirectoryContent(const MpdCom::Filelist & filelist)
{
  std::cout << "PlaylistManager::slotReceiveDirectoryContent" << std::endl;
  lvPlaylists.clear();
  lvDirectories.clear();
  if( currDir.length()!=0 ) new QListViewItem( &lvDirectories, ".." );
  lvFiles.clear();
  for(MpdCom::Filelist::iterator i=filelist.begin(); i!=filelist.end(); i++)
  {
    switch( (*i)->properties )
    {
      case MpdCom::Filelist::DIR :      new QListViewItem( &lvDirectories, (*i)->name.right( (*i)->name.length() - (*i)->name.findRev('/') - 1) );
                                        break;
      case MpdCom::Filelist::FILE :     new QListViewItem( &lvFiles, (*i)->name.right( (*i)->name.length() - (*i)->name.findRev('/') - 1) );
                                        break;
      case MpdCom::Filelist::PLAYLIST : new QListViewItem( &lvPlaylists, (*i)->name.right( (*i)->name.length() - (*i)->name.findRev('/') - 1) );
                                        break;
    }
  }
}

/*! \param item                 the item which is doubleclicked
*/
void PlaylistManager::slotDirectoryDoubleClicked(QListViewItem * item)
{
  if( item->text(0)==".." )
  {
    int tmp = currDir.findRev('/');

    if (tmp == -1)
      currDir = "";
    else
      currDir = currDir.left( tmp );
  }
  else
    currDir = currDir + (currDir.length() == 0 ? "" : "/") + item->text(0);
  mpdCom->requestDirectoryContent( currDir );
}

/*! \param item                 the item which is double clicked
*/
void PlaylistManager::slotFileDoubleClicked(QListViewItem * item)
{
  if(directPlay) mpdCom->clearPlaylist();
  mpdCom->addFile( currDir + "/" + item->text(0) );
  if(directPlay)
  {
    mpdCom->play(0);
    close();
  }
}

/*! \param item                 the item which is double clicked
*/
void PlaylistManager::slotPlaylistDoubleClicked(QListViewItem * item)
{
  mpdCom->loadPlaylist( item->text(0) );
  if(directPlay) close();
}

/*! \param event                the close event
*/
void PlaylistManager::closeEvent(QCloseEvent * event)
{
  Configuration::setGeometry("playlistmanager", geometry());
  QWidget::closeEvent(event);
}

/*! \param event                the resize event
*/
void PlaylistManager::resizeEvent(QResizeEvent * event)
{
  QWidget::resizeEvent(event);
  lvPlaylists.setColumnWidth (0, lvPlaylists.viewportSize(lvPlaylists.width(), lvPlaylists.height()).width());
  lvDirectories.setColumnWidth (0, lvDirectories.viewportSize(lvDirectories.width(), lvDirectories.height()).width());
  lvFiles.setColumnWidth (0, lvFiles.viewportSize(lvFiles.width(), lvFiles.height()).width());
}

/*! \param item                 the item on which the right click accured
    \param pos                  the position of the click
    \outside                    is equal to -1 if the click didn't accured outside the viewport
*/
void PlaylistManager::slotPlaylistRightClick(QListViewItem * item, const QPoint & pos, int outside)
{
  if(outside != -1)
  {
    QPopupMenu * pop = new QPopupMenu();
    pop->insertItem(LANG_LOAD);
    pop->insertItem(LANG_ENQUEUE);
    pop->insertItem(LANG_DELETE);

    int result = pop->exec(pos);

    if(      pop->text(result) == LANG_LOAD )   mpdCom->loadPlaylist(currDir + (currDir.length() == 0 ? "" : "/") + item->text(0));
    else if( pop->text(result) == LANG_ENQUEUE) mpdCom->loadPlaylist(currDir + (currDir.length() == 0 ? "" : "/") + item->text(0), true);
    else if( pop->text(result) == LANG_DELETE ) mpdCom->deletePlaylist(currDir + (currDir.length() == 0 ? "" : "/") + item->text(0));
    delete pop;
  }
}

/*! \param item                 the item on which the right click accured
    \param pos                  the position of the click
    \outside                    is equal to -1 if the click didn't accured outside the viewport
*/
void PlaylistManager::slotDirectoryRightClick(QListViewItem * item, const QPoint & pos, int outside)
{
  if(outside != -1)
  {
    if(item)
    if(item->text(0)!="..")
    {
      QPopupMenu * pop = new QPopupMenu();
      pop->insertItem(LANG_OPEN);
      if(mpdCom->minVersion("0.10")) pop->insertItem(LANG_ADD);

      int result = pop->exec(pos);

      if(      pop->text(result) == LANG_OPEN ) slotDirectoryDoubleClicked(item);
      else if( pop->text(result) == LANG_ADD) mpdCom->addFile(currDir + (currDir.length() == 0 ? "" : "/") + item->text(0));
      delete pop;
    }
  }
}

void PlaylistManager::slotAddSelection(void)
{
  QListViewItem * item = lvFiles.firstChild();
  if(item)
  {
    mpdCom->startCommandList();
    do
    {
      if(lvFiles.isSelected(item)) mpdCom->addFile(currDir + (currDir.length() == 0 ? "" : "/") + item->text(0));
    }while(item = item->nextSibling());
    mpdCom->endCommandList();
  }
}
